from __future__ import division, print_function

import os
from contextlib import contextmanager
from copy import copy

from sqlalchemy import create_engine
from sqlalchemy.engine.url import make_url
from sqlalchemy.orm import sessionmaker
from sqlalchemy_utils.functions import quote

from beam_nuggets.io.relational_db_api import (
    create_table,
    load_table,
    get_column_names_from_table,
    TableConfiguration,
    _create_database_tmp as create_database
)


class TestDatabase(object):
    def __init__(self, source_config):
        self._url = source_config.url
        self._SessionClass = sessionmaker(bind=create_engine(self._url))

    def init_db(self):
        create_database(self._url)

    def destroy_db(self):
        _drop_database_tmp(self._url)  # will also remove the sqlite file

    def create_table(self, name, create_table_if_missing, define_table_f):
        # create the table
        with self.session_scope() as session:
            create_table(
                session=session,
                name=name,
                table_config=TableConfiguration(
                    name=name,
                    define_table_f=define_table_f,
                    create_if_missing=create_table_if_missing
                ),
                record=None
            )

    def load_table_class(self, table_name):
        with self.session_scope() as session:
            return load_table(session, table_name)

    def write_rows(self, table_name, rows):
        with self.session_scope() as session:
            TableCls = load_table(session, table_name)
            session.add_all([TableCls(**row_dict) for row_dict in rows])

    def read_rows(self, table_name):
        with self.session_scope() as session:
            TableCls = load_table(session, table_name)
            if TableCls:
                column_names = get_column_names_from_table(TableCls)

                def to_dict(db_row):
                    return {col: getattr(db_row, col) for col in column_names}

                rows = [to_dict(db_row) for db_row in session.query(TableCls)]
            else:
                rows = []
        return rows

    @contextmanager
    def session_scope(self):
        session = self._SessionClass()
        try:
            yield session
            session.commit()
        except:
            session.rollback()
            raise
        finally:
            session.close_all()
            session.bind.dispose()


# TODO Below is a copy of sqlalchemy_utils.drop_database with a fix to
#  support dropping PostgreSQL DBs using pg8000. We should remove this tmp fix
#  when the following PR is merged:
#  https://github.com/kvesteri/sqlalchemy-utils/pull/356
def _drop_database_tmp(url):
    """Issue the appropriate DROP DATABASE statement.

    :param url: A SQLAlchemy engine URL.

    Works similar to the :ref:`create_database` method in that both url text
    and a constructed url are accepted. ::

        drop_database('postgres://postgres@localhost/name')
        drop_database(engine.url)

    """

    url = copy(make_url(url))

    database = url.database

    if url.drivername.startswith('postgres'):
        url.database = 'postgres'
    elif url.drivername.startswith('mssql'):
        url.database = 'master'
    elif not url.drivername.startswith('sqlite'):
        url.database = None

    if url.drivername == 'mssql+pyodbc':
        engine = create_engine(url, connect_args={'autocommit': True})
    elif url.drivername == 'postgresql+pg8000':
        engine = create_engine(url, isolation_level='AUTOCOMMIT')
    else:
        engine = create_engine(url)
    conn_resource = None

    if engine.dialect.name == 'sqlite' and database != ':memory:':
        if database:
            os.remove(database)

    elif engine.dialect.name == 'postgresql' and engine.driver == 'psycopg2':
        from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT

        connection = engine.connect()
        connection.connection.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)

        # Disconnect all users from the database we are dropping.
        version = connection.dialect.server_version_info
        pid_column = (
            'pid' if (version >= (9, 2)) else 'procpid'
        )
        text = '''
        SELECT pg_terminate_backend(pg_stat_activity.%(pid_column)s)
        FROM pg_stat_activity
        WHERE pg_stat_activity.datname = '%(database)s'
          AND %(pid_column)s <> pg_backend_pid();
        ''' % {'pid_column': pid_column, 'database': database}
        connection.execute(text)

        # Drop the database.
        text = 'DROP DATABASE {0}'.format(quote(connection, database))
        connection.execute(text)
        conn_resource = connection
    else:
        text = 'DROP DATABASE {0}'.format(quote(engine, database))
        conn_resource = engine.execute(text)

    if conn_resource is not None:
        conn_resource.close()
    engine.dispose()



from __future__ import division, print_function

import unittest

import testing.mysqld
import testing.postgresql

from beam_nuggets.io.relational_db import SourceConfiguration
from .database import TestDatabase


class TransformBaseTest(unittest.TestCase):
    postgres_instance = None
    mysql_instance = None

    postgres_source_config = None
    mysql_source_config = None

    source_config = None  # default for all tests

    @classmethod
    def setUpClass(cls):
        cls.postgres_source_config = cls.get_postgres_source_config()
        cls.mysql_source_config = cls.get_mysql_source_config()

        cls.source_config = (
            cls.postgres_source_config or
            cls.mysql_source_config or
            cls.get_sqlite_source_config()
        )

        print(
            '\nrunning {} tests against temp db instance: {}'
            ''.format(cls.__name__, cls.source_config.url)
        )

    @classmethod
    def get_sqlite_source_config(cls):
        return SourceConfiguration(
            drivername='sqlite',
            database='/tmp/delete_me_beam_nuggets_unittest.sqlite',
            create_if_missing=True,
        )

    @classmethod
    def get_postgres_source_config(cls):
        cls.postgres_instance = cls.connect_to_postgresql()
        if cls.postgres_instance:
            return SourceConfiguration(
                drivername='postgresql+pg8000',
                host='localhost',
                port=cls.postgres_instance.settings['port'],
                username='postgres',
                database='beam_nuggets_test_db',
                create_if_missing=True,
            )

    @classmethod
    def connect_to_postgresql(cls):
        postgresql = None
        try:
            postgresql = testing.postgresql.Postgresql()
        except Exception as e:
            print(
                '\n\nFailed to connect to local postgres instance. Error: {}'
                '\nCheck postgresql installed locally to run tests against it '
                '.\n{}\n{}'.format(
                    e,
                    'https://www.postgresql.org/download/linux/ubuntu/',
                    'https://github.com/tk0miya/testing.postgresql'
                )
            )
        return postgresql

    @classmethod
    def get_mysql_source_config(cls):
        cls.mysql_instance = cls.connect_to_mysql()
        if cls.mysql_instance:
            return SourceConfiguration(
                drivername='mysql+pymysql',
                host='localhost',
                port=cls.mysql_instance.my_cnf['port'],
                username='root',
                database='beam_nuggets_test_db',
                create_if_missing=True,
            )

    @classmethod
    def connect_to_mysql(cls):
        mysql = None
        try:
            # mysql = testing.mysqld.Mysqld(my_cnf={'skip-networking': None})
            mysql = testing.mysqld.Mysqld()
        except Exception as e:
            print(
                '\n\nFailed to connect to local mysql instance. Error: {}'
                '\nCheck mysql installed locally to run tests against it '
                '.\n{}\n{}'.format(
                    e,
                    'https://support.rackspace.com/how-to/installing-mysql-server-on-ubuntu/',
                    'https://github.com/tk0miya/testing.mysqld'
                )
            )
        return mysql

    @classmethod
    def tearDownClass(cls):
        if cls.postgres_instance:
            cls.postgres_instance.stop()
        if cls.mysql_instance:
            cls.mysql_instance.stop()

    @staticmethod
    def configure_db(source_config):
        db = TestDatabase(source_config)
        db.init_db()
        return db

    @staticmethod
    def destroy_db(db):
        # need this since some TC:s will use; I don't want to call tearDown
        # from within TC:s
        db.destroy_db()  # will, as well, delete created sqllite db file

    def setUp(self):
        self.db = self.configure_db(self.source_config)

    def tearDown(self):
        # clean DB resource
        self.destroy_db(self.db)



"""Wrapper of SqlAlchemy code used for reading from and writing to databases"""
from __future__ import division, print_function

import datetime
from copy import copy

from sqlalchemy import (
    create_engine, MetaData, Table, Column,
    Integer,
    String,
    Float,
    Boolean,
    DateTime,
    Date,
    insert as generic_insert
)
from sqlalchemy.dialects.mysql import insert as mysql_insert
from sqlalchemy.dialects.postgresql import insert as postgres_insert
from sqlalchemy.engine.url import URL, make_url
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from sqlalchemy_utils import database_exists
from sqlalchemy_utils.functions import quote


class SourceConfiguration(object):
    """Holds parameters for accessing a database.

    User to pass database access parameters to
    :class:`~beam_nuggets.io.relational_db_api.SqlAlchemyDB`.

    ``SourceConfiguration.url`` provides the database url used by SqlAlchemy to
    connect to the database.

    Args:
        drivername (str): name of the database backend. It specifies the
            target database type and the driver (DBAPI) used
            by SqlAlchemy to communicate with database. The following
            drivernames are supported by and tested on beam-nuggets:
              - postgresql+pg8000: for PostgreSQL (pg8000 is the DB driver).
              - mysql+pymysql: for MySQL.
              - sqlite: for SQLite.
            Additional drivers can be used after installing their
            corresponding python libraries. Refer to `SqlAlchemy dialects`_
            for more information on the supported databases and the
            corresponding drivers.
        host (str): database host name or IP.
        port (int): database port number.
        database (str): database name.
        username (str): database username.
        password (str): database password.
        create_if_missing (bool): If set to True, it instructs to create a
            missing database before writing to it.

    Examples:
        MySQL database ::

            from beam_nuggets.io import relational_db
            src_cnf = relational_db.SourceConfiguration(
                drivername='mysql+pymysql',
                host='localhost',
                port=37311,
                username='root',
                database='test',
                create_if_missing=True,
            )
            print(src_cnf.url)
            # mysql+pymysql://root@localhost:37311/test

        PostgreSQL database ::

            from beam_nuggets.io import relational_db
            src_cnf = relational_db.SourceConfiguration(
                drivername='postgresql',
                host='localhost',
                port=42139,
                username='postgres',
                password='pass',
                database='test'
            )
            print(src_cnf.url)
            # postgresql://postgres:pass@localhost:42139/test

        SQLite database ::

            from beam_nuggets.io import relational_db
            src_cnf = relational_db.SourceConfiguration(
                drivername='sqlite',
                database='/tmp/test_db.sqlite'
            )
            print(src_cnf.url)
            # sqlite:////tmp/test_db.sqlite

    .. _SqlAlchemy dialects: https://docs.sqlalchemy.org/en/latest/dialects/
    """

    def __init__(
        self,
        drivername,
        host=None,
        port=None,
        database=None,
        username=None,
        password=None,
        create_if_missing=False,
    ):
        self.url = URL(
            drivername=drivername,
            username=username,
            password=password,
            host=host,
            port=port,
            database=database
        )
        self.create_if_missing = create_if_missing


class TableConfiguration(object):
    """Holds parameters for a database table.
    Used to pass table parameters to :class:`SqlAlchemyDB`.

    Args:
        name (str): the table name.
        create_insert_f (function): a function that takes as input an
            SqlAlchemy table and a row record, and returns an statement for
            inserting the record into the table. The function doesn't
            execute the insert statement. If not specified, the following
            default implementations are used:
              - :func:`create_upsert_mysql` for MySQL tables.
              - :func:`create_upsert_postgres` for PostgreSQL tables.
              - :func:`create_insert` for other databases.
            As a mechanism to recover from failures, beam runners will
            attempt to apply a transform multiple times on the same data;
            because of that it is recommended to implement idempotent writes
            (e.g. :func:`create_upsert_mysql` and
            :func:`create_upsert_postgres`)
            to avoid data inconsistency issues arising from this beam behavior.
            The function has the following signature:
            (``sqlalchemy.sql.schema.Table``, ``dict``) ->
            ``sqlalchemy.sql.dml.Insert``.
        create_if_missing (bool): if set to True and the table is missing
            :class:`SqlAlchemyDB` will create the table. See below notes on new
            table creation. See below note how this is used when creating
            new tables.
        primary_key_columns (list): list of column names to be used as
            primary key (if multiple columns are specified, a composite key
            is created), when creating the table. See below notes on new table
            creation.
        define_table_f (function): A function for defining an SqlAlchemy
            table (the function doesn't create the table); the definition is
            used when creating the table. The function has the following
            signature: (``Sqlalchemy.Metadata``) -> ``sqlalchemy.Table``. See
            below notes on how this is used when creating missing tables.
            See this `define table tutorial`_ for how to implement the
            function.

    Notes:
        When attempting to write to a missing database table,
        :class:`SqlAlchemyDB` will handle the situation based on the values
        ``create_if_missing``, ``primary_key_columns`` and ``define_table_f``
        of the passed :class:`TableConfiguration`, as follows:
          - If the table is missing and ``create_if_missing`` is set to
            False (default), :class:`SqlAlchemyDB` will raise an exception.
          - Only when the target table is missing and ``create_if_missing``
            is set to True, table creation is attempted. This is the assumed
            state for all the following cases.
          - If ``define_table_f`` is specified, a new table will be created
            using the table definition returned by ``define_table_f``,
            irrespective of ``primary_key_columns``.
          - If ``primary_key_columns`` is specified and ``define_table_f``
            is None, a new table will be created using the columns specified
            in ``primary_key_columns`` as the primary key. The full column
            list and their types are inferred automatically using the first
            record to be written. See :func:`infer_db_type` for information on
            the how the database column types are inferred from the python
            types. If ``primary_key_columns`` is also None, an auto_increment
            Integer column will be created and used as primary key this is
            done as some databases require a primary key to be specified when
            creating tables.

    Examples:
        A configuration for creating the table if missing using the specified
        columns to create the primary key. ::

            from beam_nuggets.io import relational_db

            table_config = relational_db.TableConfiguration(
                name='students',
                create_if_missing=True,
                primary_key_columns=['id']
            )


        A configuration for creating the table if missing using the specified
        definition. ::

            from sqlalchemy import Table, Integer, String, Column
            from beam_nuggets.io import relational_db

            table_name = 'students'

            def define_students_table(metadata):
                return Table(
                    table_name, metadata,
                    Column(ID, Integer, primary_key=True),
                    Column(NAME, String(100)),
                    Column(AGE, Integer)
                )

            table_config = relational_db.TableConfiguration(
                name=table_name,
                create_if_missing=True,
                define_table_f=define_students_table
            )

    .. _define table tutorial:
       https://docs.sqlalchemy.org/en/latest/core/tutorial.html#define-and
       -create-tables
    """

    def __init__(
        self,
        name,
        define_table_f=None,
        create_if_missing=False,
        primary_key_columns=None,
        create_insert_f=None
    ):
        self.name = name
        self.define_table_f = define_table_f
        self.create_table_if_missing = create_if_missing
        self.primary_key_column_names = primary_key_columns or []
        self.create_insert_f = create_insert_f


class SqlAlchemyDB(object):
    """Provides functionality to read and write from and to relational DBs.
    It uses SqlAlchemy.

    Args:
        source_config (SourceConfiguration): Information for accessing the
            target database.
    """

    def __init__(self, source_config):
        self._source = source_config

        self._SessionClass = sessionmaker(bind=create_engine(self._source.url))
        self._session = None  # will be set in self.start_session()

        self._name_to_table = {}  # tables metadata cache

    def start_session(self):
        create_if_missing = self._source.create_if_missing
        is_database_missing = lambda: not database_exists(self._source.url)
        if create_if_missing and is_database_missing():
            _create_database_tmp(self._source.url)
        self._session = self._SessionClass()

    def close_session(self):
        self._session.close_all()
        self._session.bind.dispose()
        self._session = None

    def read(self, table_name):
        table = self._open_table_for_read(table_name)
        for record in table.records(self._session):
            yield record

    def write_record(self, table_config, record_dict):
        """Writes a single record to the specified table.

        Args:
            table_config (TableConfiguration): specifies the target table,
                 how data should inserted and how to create it if it was
                 missing. See :class:`TableConfiguration` notes on table
                 creation.
            record_dict (dict): the record to be written
        """
        table = self._open_table_for_write(table_config, record_dict)
        table.write_record(
            session=self._session,
            create_insert_f=self._get_create_insert_f(table_config),
            record_dict=record_dict
        )

    def _get_create_insert_f(self, table_config):
        create_insert_f = table_config.create_insert_f
        if not create_insert_f:
            if 'postgresql' in self._source.url.drivername:
                create_insert_f = create_upsert_postgres
            elif 'mysql' in self._source.url.drivername:
                create_insert_f = create_upsert_mysql
            else:
                create_insert_f = create_insert
        return create_insert_f

    def _open_table_for_read(self, name):
        return self._open_table(
            name=name,
            get_table_f=load_table
        )

    def _open_table_for_write(self, table_config, record):
        return self._open_table(
            name=table_config.name,
            get_table_f=create_table,
            table_config=table_config,
            record=record
        )

    def _open_table(self, name, get_table_f, **get_table_f_params):
        table = self._name_to_table.get(name, None)
        if not table:
            self._name_to_table[name] = (
                self._get_table(name, get_table_f, **get_table_f_params)
            )
            table = self._name_to_table[name]
        return table

    def _get_table(self, name, get_table_f, **get_table_f_params):
        table_class = get_table_f(self._session, name, **get_table_f_params)
        if table_class:
            table = _Table(table_class=table_class, name=name)
        else:
            raise SqlAlchemyDbException('Failed to get table {}'.format(name))
        return table


class SqlAlchemyDbException(Exception):
    pass


class _Table(object):
    def __init__(self, table_class, name):
        self._Class = table_class
        self._sqlalchemy_table = table_class.__table__
        self.name = name
        self._column_names = get_column_names_from_table(table_class)

    def records(self, session):
        for record in session.query(self._Class):
            yield self._from_db_record(record)

    def write_record(self, session, create_insert_f, record_dict):
        try:
            insert_stmt = create_insert_f(
                table=self._sqlalchemy_table,
                record=record_dict
            )
            session.execute(insert_stmt)
            session.commit()
        except:
            session.rollback()
            session.close_all()
            session.bind.dispose()
            raise

    def _to_db_record(self, record_dict):
        return self._Class(**record_dict)

    def _from_db_record(self, db_record):
        return {col: getattr(db_record, col) for col in self._column_names}


def load_table(session, name):
    table_class = None
    engine = session.bind
    if engine.dialect.has_table(engine, name):
        metadata = MetaData(bind=engine)
        table_class = create_table_class(Table(name, metadata, autoload=True))
    return table_class


def create_table(session, name, table_config, record):
    # Attempt to load from the DB
    table_class = load_table(session, name)

    if not table_class and table_config.create_table_if_missing:
        define_table_f = (
            table_config.define_table_f or
            _get_default_define_f(
                record=record,
                name=name,
                primary_key_column_names=table_config.primary_key_column_names,
                drivername=session.bind.url.drivername,
            )
        )
        metadata = MetaData(bind=session.bind)
        sqlalchemy_table = define_table_f(metadata)
        metadata.create_all()
        table_class = create_table_class(sqlalchemy_table)

    return table_class


def create_table_class(sqlalchemy_table):
    class TableClass(declarative_base()):
        __table__ = sqlalchemy_table

    return TableClass


def _get_default_define_f(record, name, primary_key_column_names, drivername):
    def define_table(metadata):
        """Defines an SqlAlchemy database table and adds it to the specified
        metadata object.

        Args:
            metadata (sqlalchemy.Metadata): database metadata.

        Returns:
            sqlalchemy.Table: A database table added to the passed metadata.
        """
        columns = _columns_from_sample_record(
            record=record,
            primary_key_column_names=primary_key_column_names,
            drivername=drivername
        )
        return Table(name, metadata, *columns)

    return define_table


def _columns_from_sample_record(record, primary_key_column_names, drivername):
    if len(primary_key_column_names) > 0:
        primary_key_columns = [
            Column(
                col, infer_db_type(record[col], drivername), primary_key=True
            )
            for col in primary_key_column_names
        ]
        other_columns = [
            Column(col, infer_db_type(value, drivername))
            for col, value in record.iteritems()
            if col not in primary_key_column_names
        ]
    else:
        pri_col_name = 'id'
        while pri_col_name in record.keys():
            pri_col_name += '_'
        primary_key_columns = [Column(pri_col_name, Integer, primary_key=True)]
        other_columns = [
            Column(col, infer_db_type(value, drivername))
            for col, value in record.iteritems()
        ]
    return primary_key_columns + other_columns


def create_insert(table, record):
    """Creates a statement for inserting the passed record to the passed table.
    The created statement is not executed by this function.

    For information on creating insert and update statements Refer to these
    SqlAlchemy `documentation`_ and `tutorial`_.

    Args:
        table (sqlalchemy.sql.schema.Table): database table metadata.
        record (dict): a data record, corresponding to one row, to be inserted.

    Returns:
        sqlalchemy.sql.dml.Insert: a statement for inserting the passed
        record to the specified table.


    .. _documentation: https://docs.sqlalchemy.org/en/latest/core/dml.html
    .. _tutorial: https://docs.sqlalchemy.org/en/latest/core/tutorial.html#insert-expressions
    """
    return generic_insert(table).values(record)


def create_upsert_postgres(table, record):
    """Creates a statement for inserting the passed record to the passed
    table; if the record already exists, the existing record will be updated.
    This uses PostgreSQL `on_conflict_do_update` (hence upsert), and that
    why the returned statement is just valid for PostgreSQL tables. Refer to
    this `SqlAlchemy PostgreSQL documentation`_ for more information.

    The created statement is not executed by this function.

    Args:
        table (sqlalchemy.sql.schema.Table): database table metadata.
        record (dict): a data record, corresponding to one row, to be inserted.

    Returns:
        sqlalchemy.sql.dml.Insert: a statement for inserting the passed
        record to the specified table.

    .. _SqlAlchemy PostgreSQL documentation:
       https://docs.sqlalchemy.org/en/latest/dialects/postgresql.html#insert-on-conflict-upsert
    """
    insert_stmt = postgres_insert(table).values(record)
    return insert_stmt.on_conflict_do_update(
        index_elements=[col for col in table.primary_key],
        set_=record
    )


def create_upsert_mysql(table, record):
    """Creates a statement for inserting the passed record to the passed
    table; if the record already exists, the existing record will be updated.
    This uses MySQL `on_duplicate_key_update` (hence upsert), and that
    why the returned statement is valid only for MySQL tables. Refer to this
    `SqlAlchemy MySQL documentation`_ for more information.

    The created statement is not executed by this function.

    Args:
        table (sqlalchemy.sql.schema.Table): database table metadata.
        record (dict): a data record, corresponding to one row, to be inserted.

    Returns:
        sqlalchemy.sql.dml.Insert: a statement for inserting the passed
        record to the specified table.

    .. _SqlAlchemy MySQL documentation:
       https://docs.sqlalchemy.org/en/latest/dialects/mysql.html#mysql-inser-on-duplicate-key-update
    """
    insert_stmt = mysql_insert(table).values(record)
    return insert_stmt.on_duplicate_key_update(**record)
    # passing dict, i.e. ...update(record), isn't working


def get_column_names_from_table(table_class):
    return [col.name for col in table_class.__table__.columns]


def infer_db_type(val, drivername):
    """Infer a database column type for storing values of the same type as the
    passed variable val in a database identified by drivername.

    Column types are inferred based on the following mapping:

    +-----------------------+--------------------------------------------------------------------------------------------------------+
    | Python type           | Column type                                                                                            |
    +=======================+========================================================================================================+
    | ``bool``              | ``Boolean``                                                                                            |
    +-----------------------+--------------------------------------------------------------------------------------------------------+
    | ``<number>``          | ``Float`` (All python numbers are mapped to Float columns)                                             |
    +-----------------------+--------------------------------------------------------------------------------------------------------+
    | ``datetime.datetime`` | ``DateTime``                                                                                           |
    +-----------------------+--------------------------------------------------------------------------------------------------------+
    | ``datetime.date``     | ``Date``                                                                                               |
    +-----------------------+--------------------------------------------------------------------------------------------------------+
    | ``<default>``         | ``String`` for PostgreSQL and SQLite and String(:const:`VARCHAR_DEFAULT_COL_SIZE`) for other databases |
    +-----------------------+--------------------------------------------------------------------------------------------------------+

    Args:
        val (object): value used to infer the database column type.
        drivername: specifies the database type and driver used for
            connecting to the database (the driver information isn't used to
            infer the column type).

    Returns:
        object: one of sqlalchemy column types.
    """
    for is_type_f, db_type in PYTHON_TO_DB_TYPE:
        if is_type_f(val):
            return db_type
    return (
        String if _does_support_varchar(drivername) else
        String(VARCHAR_DEFAULT_COL_SIZE)
        # It seems only PostgreSQL and SQLite support String columns with
        # not specified length.
    )


VARCHAR_DEFAULT_COL_SIZE = 50


def _does_support_varchar(drivername):
    return 'postgresql' in drivername or 'sqlite' in drivername


def _is_number(x):
    try:
        _ = x + 1
    except:
        return False
    return not hasattr(x, '__len__')


PYTHON_TO_DB_TYPE = [
    # Order matters!
    (lambda x: isinstance(x, bool), Boolean),
    (_is_number, Float),
    (lambda x: isinstance(x, datetime.datetime), DateTime),
    (lambda x: isinstance(x, datetime.date), Date),
]


# TODO Below is a copy of sqlalchemy_utils.create_database with a fix to
#  support creating PostgreSQL DBs using pg8000. We should remove this tmp fix
#  when the following PR is merged:
#  https://github.com/kvesteri/sqlalchemy-utils/pull/356
def _create_database_tmp(url, encoding='utf8', template=None):
    """Issue the appropriate CREATE DATABASE statement.

    :param url: A SQLAlchemy engine URL.
    :param encoding: The encoding to create the database as.
    :param template:
        The name of the template from which to create the new database. At the
        moment only supported by PostgreSQL driver.

    To create a database, you can pass a simple URL that would have
    been passed to ``create_engine``. ::

        create_database('postgres://postgres@localhost/name')

    You may also pass the url from an existing engine. ::

        create_database(engine.url)

    Has full support for mysql, postgres, and sqlite. In theory,
    other database engines should be supported.
    """

    url = copy(make_url(url))

    database = url.database

    if url.drivername.startswith('postgres'):
        url.database = 'postgres'
    elif url.drivername.startswith('mssql'):
        url.database = 'master'
    elif not url.drivername.startswith('sqlite'):
        url.database = None

    if url.drivername == 'mssql+pyodbc':
        engine = create_engine(url, connect_args={'autocommit': True})
    elif url.drivername == 'postgresql+pg8000':
        engine = create_engine(url, isolation_level='AUTOCOMMIT')
    else:
        engine = create_engine(url)
    result_proxy = None

    if engine.dialect.name == 'postgresql':
        if engine.driver == 'psycopg2':
            from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT
            engine.raw_connection().set_isolation_level(
                ISOLATION_LEVEL_AUTOCOMMIT
            )

        if not template:
            template = 'template1'

        text = "CREATE DATABASE {0} ENCODING '{1}' TEMPLATE {2}".format(
            quote(engine, database),
            encoding,
            quote(engine, template)
        )
        result_proxy = engine.execute(text)

    elif engine.dialect.name == 'mysql':
        text = "CREATE DATABASE {0} CHARACTER SET = '{1}'".format(
            quote(engine, database),
            encoding
        )
        result_proxy = engine.execute(text)

    elif engine.dialect.name == 'sqlite' and database != ':memory:':
        if database:
            engine.execute("CREATE TABLE DB(id int);")
            engine.execute("DROP TABLE DB;")

    else:
        text = 'CREATE DATABASE {0}'.format(quote(engine, database))
        result_proxy = engine.execute(text)

    if result_proxy is not None:
        result_proxy.close()
    engine.dispose()



from __future__ import division, print_function

import csv

import apache_beam as beam


class Read(beam.PTransform):
    """A :class:`~apache_beam.transforms.ptransform.PTransform` for reading
    csv files.

    It outputs a :class:`~apache_beam.pvalue.PCollection` of
    ``dict:s``, each corresponding to a row in the csv file.

    Args:
        csv_path (str): csv file path.

    Examples:
        Reading content of a csv file. ::

            import apache_beam as beam
            from apache_beam.options.pipeline_options import PipelineOptions
            from beam_nuggets.io import csvio

            path_to_csv = '/path/to/students.csv'
            with beam.Pipeline(options=PipelineOptions()) as p:
                students = p | "Reading students records" >> csvio.Read(path_to_csv)
                students | 'Writing to stdout' >> beam.Map(print)

        The output will be something like ::

            {'lastName': 'Norvell', 'firstName': 'Andrel', 'level': '10'}
            {'lastName': 'Proudfoot', 'firstName': 'Dinorah', 'level': '8'}
            {'lastName': 'Plotkin', 'firstName': 'Trulal', 'level': '14'}

    """

    def __init__(self, csv_path, *args, **kwargs):
        super(Read, self).__init__(*args, **kwargs)
        self._csv_path = csv_path

    def expand(self, pcoll):
        return pcoll | beam.io.Read(_CsvSource(self._csv_path))


class _CsvSource(beam.io.filebasedsource.FileBasedSource):
    def read_records(self, file_name, range_tracker):
        # FIXME handle concurrent read
        self._file = self.open_file(file_name)

        for rec in csv.DictReader(self._file):
            yield rec



""":class:`~apache_beam.transforms.ptransform.PTransform` s for reading from
and writing to relational databases. SQLAlchemy_ is used for interfacing the
databases.

.. _SQLAlchemy: https://www.sqlalchemy.org/
"""

from __future__ import division, print_function

from apache_beam import PTransform, Create, ParDo, DoFn

from beam_nuggets.io.relational_db_api import (
    SqlAlchemyDB,
    SourceConfiguration,
    TableConfiguration
)

# It is intended SourceConfiguration and TableConfiguration are imported
# from this module by the library users. Below is to make sure they've been
# imported to this module.
assert SourceConfiguration is not None
assert TableConfiguration is not None


class Read(PTransform):
    """A :class:`~apache_beam.transforms.ptransform.PTransform` for reading
    tables on relational databases.

    It outputs a :class:`~apache_beam.pvalue.PCollection` of
    ``dict:s``, each corresponding to a row in the target database table.

    Args:
        source_config (SourceConfiguration): specifies the target database.
        table_name (str): the name of the table to be read.

    Examples:
        Reading from a table on a postgres database. ::

            import apache_beam as beam
            from apache_beam.options.pipeline_options import PipelineOptions
            from beam_nuggets.io import relational_db

            source_config = relational_db.SourceConfiguration(
                drivername='postgresql',
                host='localhost',
                port=5432,
                username='postgres',
                password='password',
                database='calendar',
            )
            table_name = 'months'

            with beam.Pipeline(options=PipelineOptions()) as p:
                records = p | "Reading records from db" >> relational_db.Read(
                    source_config=source_config,
                    table_name=table_name,
                )
                records | 'Writing to stdout' >> beam.Map(print)

        The output will be something like ::

            {'name': 'Jan', 'num': 1}
            {'name': 'Feb', 'num': 2}

        Where "name" and "num" are the column names.
    """

    def __init__(self, source_config, table_name, *args, **kwargs):
        super(Read, self).__init__(*args, **kwargs)
        self._read_args = dict(
            source_config=source_config,
            table_name=table_name
        )

    def expand(self, pcoll):
        return (
            pcoll
            | Create([self._read_args])
            | ParDo(_ReadFromRelationalDBFn())
        )


class _ReadFromRelationalDBFn(DoFn):
    def process(self, element):
        db_args = dict(element)
        table_name = db_args.pop('table_name')
        db = SqlAlchemyDB(**db_args)
        db.start_session()
        try:
            for record in db.read(table_name):
                yield record
        except:
            raise
        finally:
            db.close_session()


class Write(PTransform):
    """A :class:`~apache_beam.transforms.ptransform.PTransform` for writing
    to tables on relational databases.

    Args:
        source_config (SourceConfiguration): specifies the target database. If
            ``source_config.create_if_missing`` is set to True, the target
            database will be created if it was missing.
        table_config (TableConfiguration): specifies the target table, as well
            as other optional parameters:
              - option for specifying custom insert statements.
              - options for creating the target table if it was missing.
            See :class:`~beam_nuggets.io.relational_db_api.TableConfiguration`
            for details.

    Examples:
        Writing to postgres database table. ::

            import apache_beam as beam
            from apache_beam.options.pipeline_options import PipelineOptions
            from beam_nuggets.io import relational_db

            records = [
                {'name': 'Jan', 'num': 1},
                {'name': 'Feb', 'num': 2},
            ]

            source_config = relational_db.SourceConfiguration(
                drivername='postgresql',
                host='localhost',
                port=5432,
                username='postgres',
                password='password',
                database='calendar',
            )

            table_config = relational_db.TableConfiguration(
                name='months',
                create_if_missing=True,
                primary_key_columns=['num']
            )

            with beam.Pipeline(options=PipelineOptions()) as p:
                months = p | "Reading month records" >> beam.Create(records)
                months | 'Writing to DB table' >> relational_db.Write(
                    source_config=source_config,
                    table_config=table_config
                )

    """

    def __init__(self, source_config, table_config, *args, **kwargs):
        super(Write, self).__init__(*args, **kwargs)
        self.source_config = source_config
        self.table_config = table_config

    def expand(self, pcoll):
        return pcoll | ParDo(_WriteToRelationalDBFn(
            source_config=self.source_config,
            table_config=self.table_config
        ))


class _WriteToRelationalDBFn(DoFn):
    def __init__(self, source_config, table_config, *args, **kwargs):
        super(_WriteToRelationalDBFn, self).__init__(*args, **kwargs)
        self.source_config = source_config
        self.table_config = table_config

    def start_bundle(self):
        self._db = SqlAlchemyDB(self.source_config)
        self._db.start_session()

    def process(self, element):
        assert isinstance(element, dict)
        self._db.write_record(self.table_config, element)

    def finish_bundle(self):
        self._db.close_session()



from __future__ import division, print_function

import apache_beam as beam


class SelectFromNestedDict(beam.DoFn):
    """A :class:`~apache_beam.DoFn` for selecting subsets of records formed of
    nested dictionaries.

    Args:
        keys (list): list of dictionary keys to be selected. Each key is a
            string formed of dot separated words, each is used for selecting
            from a dict in the nested dicts. The order of the word in the
            "dot.separated.string" correspond to the dict level to select from.
            For instance, if the input record is ``{'a': {'b': 3, 'c': {'d': 2} }}``
            and keys is ``['a.b', 'a.c.d']``, the transform output will
            ``{'a_b': 3, 'a_c_d': 2} or {'b': 3, 'd': 2}`` depending on
            ``deepest_key_as_name `` below.
        deepest_key_as_name (bool): if set to True, the deepest selected fields
            keys will be used as names for the output dict keys.

    Examples:
        Select from records formed of nested dicts. ::

            import apache_beam as beam
            from apache_beam.options.pipeline_options import PipelineOptions
            from beam_nuggets.transforms import SelectFromNestedDict

            inputs = [
                {
                    'name': {'first': 'Star', 'second': 'Light'},
                    'address': {'st': 'Jupiter', 'flat': 3},
                    'email': 's@l.no'
                },
                {
                    'name': {'first': 'Mark', 'second': 'Sight'},
                    'address': {'st': 'Loon', 'flat': 5},
                    'email': 'm@s.no'
                }
            ]
            with beam.Pipeline(options=PipelineOptions()) as p:
                nested = p | "Reading nested dicts" >> beam.Create(inputs)
                transformed = nested | "filtering" >> beam.ParDo(SelectFromNestedDict(
                    keys=['name.first', 'address.st', 'email'],
                    # deepest_key_as_name=True,
                ))
                transformed | 'Writing to stdout' >> beam.Map(print)

        The output will be something like:

            {'address_st': 'Jupiter', 'name_first': 'Star', 'email': 's@l.no'}
            {'address_st': 'Loon', 'name_first': 'Mark', 'email': 'm@s.no'}

    """

    def __init__(self, keys, deepest_key_as_name=False, *args, **kwargs):
        super(SelectFromNestedDict, self).__init__(*args, **kwargs)
        self._compiled_keys = self._compile_keys(keys, deepest_key_as_name)

    def process(self, element):
        """
        Args:
            element(dict):
        """

        yield {
            out_key: self._retrieve(nested_keys, element)
            for nested_keys, out_key in self._compiled_keys
        }

    @staticmethod
    def _retrieve(nested_keys, element):
        for key in nested_keys:
            element = element[key]
        return element

    @staticmethod
    def _compile_keys(keys, deepest_key_as_name):
        def _get_out_dict_key(nested_keys):
            if deepest_key_as_name:
                return nested_keys[-1]
            else:
                return '_'.join(nested_keys)

        return [
            (
                nested_keys,  # ['a', 'b'] used for retrieving nested values
                _get_out_dict_key(nested_keys),  # 'a_b' or 'b'
            )
            for nested_keys in [key.split('.') for key in keys]  # ['a.b']
        ]



from __future__ import division, print_function

from setuptools import setup, find_packages

VERSION = '0.11.0'

REQUIRED_PACKAGES = [
    'apache-beam>=2.8.0,<3.0.0',
    'SQLAlchemy>=1.2.14,<2.0.0',
    'sqlalchemy-utils>=0.33.8,<0.34',
    # Below are drivers for connection to specific DBs
    'pg8000>=1.12.4,<2.0.0',
    'PyMySQL>=0.9.3,<2.0.0'
]

REQUIRED_PACKAGES_TEST = [
    'nose>=1.3.7,<2.0.0',
    'testing.postgresql>=1.3.0,<2.0.0',
    'testing.mysqld>=1.4.0,<2.0.0',
    'numpy>=1.15.4,<2.0.0',
    'pandas>=0.23.4,<0.24'
]

REQUIRED_PACKAGES_DOCS = [
    'Sphinx>=1.8.3,<2.0.0',
    'sphinx_rtd_theme>=0.4.2,<2.0.0'
]

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='beam-nuggets',
    version=VERSION,
    author='Mohamed Haseeb',
    author_email='m@mohaseeb.com',
    description='Collection of transforms for the Apache beam python SDK.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mohaseeb/beam-nuggets",
    install_requires=REQUIRED_PACKAGES,
    extras_require={'dev': REQUIRED_PACKAGES_TEST + REQUIRED_PACKAGES_DOCS},
    packages=find_packages(exclude=("test", "tests")),
    classifiers=[
        "Programming Language :: Python",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
)
